#define  thd_t  (ithd % _NBT)
#define  thd_z  (ithd /(_NBT) % _NBZ)
#define  thd_y  (ithd /(_NBT*_NBZ) % _NBY)
#define  thd_x  (ithd /(_NBT*_NBZ*_NBY) % _NBX)
#define  blk_t  (iblk % _NGT)
#define  blk_z  (iblk /(_NGT) % _NGZ)
#define  blk_y  (iblk /(_NGT*_NGZ) % _NGY)
#define  blk_x  (iblk /(_NGT*_NGZ*_NGY) % _NGX)

static __global__ void 
assign_mult_clover_term_kernel(      cuhqfield       *fye,
                               const cuhqfield        *ye,
                               const int             *ieo,
                               const cuhclvfield *fclinve)
////////////////////////////////////////////////////////////////////////////////
// Multiply inverse clover term : fye = Fee ye
//
//     fye : even/odd site fermion (ieo=0/1) (output)
//      ye : odd/even site fermion (ieo=0/1) (input)
//     ieo : operate even/odd site
//           ieo=0 for even site
//           ieo=1 for odd  site
// fclinve : even/odd site inverse clover term (ieo=0/1)
////////////////////////////////////////////////////////////////////////////////
{
  int ithd = threadIdx.x;
  int iblk = blockIdx.x;

  __align__(16) float  ytr[CLSP],yti[CLSP];

  *(float4 *)(ytr+0) = (*ye).b[iblk].yri[0].c[ithd];
  *(float4 *)(ytr+4) = (*ye).b[iblk].yri[1].c[ithd];
  *(float4 *)(ytr+8) = (*ye).b[iblk].yri[2].c[ithd];
  *(float4 *)(yti+0) = (*ye).b[iblk].yri[3].c[ithd];
  *(float4 *)(yti+4) = (*ye).b[iblk].yri[4].c[ithd];
  *(float4 *)(yti+8) = (*ye).b[iblk].yri[5].c[ithd];

  cu_mult_clover_term(&((*fclinve).b[iblk]),&ithd,ytr,yti);

#ifdef _SF
  set_sf_boundary_q(ytr,yti,ithd,iblk,ieo);
#endif

  (*fye).b[iblk].yri[0].c[ithd] = *(float4 *)(ytr+0);
  (*fye).b[iblk].yri[1].c[ithd] = *(float4 *)(ytr+4);
  (*fye).b[iblk].yri[2].c[ithd] = *(float4 *)(ytr+8);
  (*fye).b[iblk].yri[3].c[ithd] = *(float4 *)(yti+0);
  (*fye).b[iblk].yri[4].c[ithd] = *(float4 *)(yti+4);
  (*fye).b[iblk].yri[5].c[ithd] = *(float4 *)(yti+8);
}

static void 
assign_mult_clover_term(      cuhqfield   *fye,
                        const cuhqfield   *ye,
                        const int         *ieo,
                        const cuhclvfield *fclinve)
{
  assign_mult_clover_term_kernel<<<BLOCKS,THREADS>>>(fye,ye,ieo,fclinve);
  flop_mult_count += MULT_CLOVER;
  CUT_CHECK_ERROR("assign_mult_clover_term kernel execution faild");
}

#undef  thd_t
#undef  thd_z
#undef  thd_y
#undef  thd_x
#undef  blk_t
#undef  blk_z
#undef  blk_y
#undef  blk_x
