#include <config.h>

!
!= Hadron measuement controll routines
!
!== version
!
! $Id: main_hadron_class.F90,v 1.3 2011/10/31 08:11:59 ishikawa Exp $
!
module main_hadron_class
!
! This contains Hadron measurement controll routines.
!
!
!
  use config_profile_class
  use timer_class
  use comlib
  use lattice_class
  use hmc_identity
  use logfile_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use quark_prop_class
  use wavefunc_class
  implicit none
  private

  public :: main_hadron_obj
  public :: new,read,print,delete

  type main_hadron_obj
  !
  ! Hadron measurement parameters
  !
    type(vfield_gluon_wg),           pointer :: u => NULL()
    type(qprop_params),              pointer :: quark_parameter(:) => NULL()
    type(src_wavefunc_obj),          pointer :: wavefunc(:) => NULL()
    type(config_profile_obj),        pointer :: config_prof => NULL()
    character(len=CHARLEN) :: config_name,config_path,fname,input_fname,cwd,job_id,prog_version
    character(len=CHARLEN) :: cdate,ctime,czone,comment,str
    character(len=CHARLEN) :: hadron_path
    integer :: num_quark, num_wave
    integer :: ini_traj,end_traj,traj_skip,traj
    type(logfile), pointer :: solver_log => NULL()
    type(timer) :: run_time
  end type

  interface new
    module procedure new_main
  end interface

  interface delete
    module procedure delete_main
  end interface

  interface read
    module procedure read_main
    module procedure read_config
  end interface

  interface print
    module procedure print_main
  end interface

contains

subroutine new_main(this)
!
! initialize hadron measurement
!
  implicit none
  type(main_hadron_obj), intent(inout) :: this
  type(lattice_world)   :: lattice

  character(len=8)  :: cdate
  character(len=10) :: ctime
  character(len=CHARLEN) :: fname

  call new(lattice)

  call new(this%run_time)
  call tic(this%run_time)

  this%input_fname=repeat(' ',len(this%input_fname))
  this%cwd        =repeat(' ',len(this%cwd))
  this%job_id     =repeat(' ',len(this%job_id))
  this%prog_version =repeat(' ',len(this%prog_version))

  call DATE_AND_TIME(date=cdate,time=ctime)
  write(this%job_id,'(A,1X,A)')TRIM(ADJUSTL(cdate)),TRIM(ADJUSTL(ctime))
  
  this%prog_version = TRIM(hmc_vers)

  if (nodeid == 0) then
    call GET_COMMAND_ARGUMENT(1,this%input_fname)
    call getcwd(this%cwd)
  endif

  if (nodeid == 0) then
    write(*,'(" DATE AND TIME :",A)')TRIM(this%job_id)
    write(*,'(A)')TRIM(this%prog_version)
  endif

  if (NPU > 1) then
    call comlib_bcast(this%input_fname,0)
    call comlib_bcast(this%cwd,0)
    call comlib_bcast(this%job_id,0)
  endif

  allocate(this%u)
  call new(this%u)

  allocate(this%config_prof)

  return
end subroutine

subroutine delete_main(this)
!
! delete hadron measurement
!
  implicit none
  type(main_hadron_obj), intent(inout) :: this
  type(lattice_world)   :: lattice
  integer :: iqk
  real(DP) :: etime

  do iqk=1,this%num_wave
    call delete(this%wavefunc(iqk))
  enddo

  do iqk=1,this%num_quark
    call delete(this%quark_parameter(iqk))
  enddo

  if (associated(this%solver_log)) then
     call delete(this%solver_log)
     deallocate(this%solver_log)
     this%solver_log => NULL()
  endif

  deallocate(this%config_prof)
  deallocate(this%wavefunc)
  deallocate(this%quark_parameter)

  call toc(this%run_time)
  etime = get_elapse(this%run_time)
  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'(" Hadron ETime:",E24.16)')etime
  endif
  call delete(lattice)

  return
end subroutine

subroutine read_main(this,iout)
!
! read Hadron measurement parameters
!
  implicit none
  type(main_hadron_obj), intent(inout) :: this
  integer,               intent(in) :: iout
  integer :: iqk,iwf
  character(len=CHARLEN) :: char,solver_log_fname

  if (nodeid == 0) then
    open(iout,file=this%input_fname,form='formatted',action='read',status='old')
    read(iout,*)
    read(iout,*)this%ini_traj,this%end_traj,this%traj_skip
    read(iout,'(A)')this%config_path
    read(iout,'(A)')this%config_name
    read(iout,*)
    read(iout,'(A)')this%hadron_path
    read(iout,*)
    read(iout,*)this%num_wave
  endif

  if (NPU > 1) then
    call comlib_bcast(this%ini_traj,0)
    call comlib_bcast(this%end_traj,0)
    call comlib_bcast(this%traj_skip,0)
    call comlib_bcast(this%config_path,0)
    call comlib_bcast(this%config_name,0)
    call comlib_bcast(this%hadron_path,0)
    call comlib_bcast(this%num_wave,0)
  endif

  allocate(this%solver_log)
  write(char,'(I6.6,"-",I6.6)')this%ini_traj,this%end_traj
  solver_log_fname = "quark_solver_log."//TRIM(ADJUSTL(char))
  call new(this%solver_log,solver_log_fname)

  allocate(this%wavefunc(this%num_wave))
  do iqk=1,this%num_wave
    call new(this%wavefunc(iqk))
  enddo

  do iwf=1,this%num_wave
    if (nodeid == 0) then
      read(iout,*)
    endif
    call read(this%wavefunc(iwf),iout)
  enddo

  if (nodeid == 0) then
    read(iout,*)
    read(iout,*)this%num_quark
  endif

  if (NPU > 1) then
    call comlib_bcast(this%num_quark,0)
  endif

  allocate(this%quark_parameter(this%num_quark))
  do iqk=1,this%num_quark
    call new(this%quark_parameter(iqk))
    this%quark_parameter(iqk)%solver_log => this%solver_log
  enddo

  do iqk=1,this%num_quark
    if (nodeid == 0) then
      read(iout,*)
    endif
    call read(this%quark_parameter(iqk),iout)
  enddo

  return
end subroutine

#include "read_config.h90"

subroutine print_main(this)
!
! print out hadron measurement parameters
!
  implicit none
  type(main_hadron_obj), intent(inout) :: this
  type(lattice_world) :: lattice
  integer :: iqk

  call print(lattice)

  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'("   Hadron Save path :",A)')TRIM(this%hadron_path)
    write(*,'(80("="))')
  endif
  
  return
end subroutine

end module
